//This program is licensed under the GPLv3; see license.txt for more info.
//Copyright 2008 Sprite_tm

#include "io.h"
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#define LED1 PD0
#define LED2 PD1
#define LED3 PD2
#define LED4 PD3
#define LED5 PD4
#define LED6 PB6
#define LED7 PB7
#define LED8 PD5
#define LED9 PD6

#define FOTO_MEAS	PC0

volatile char ledIntensity[9];
volatile short pressedButtons;
volatile char calButtons;


//Init all io-ports plus the led timer
void initIo(void) {
    DDRD=(1<<LED1)|(1<<LED2)|(1<<LED3)|(1<<LED4)|(1<<LED5)|(1<<LED8)|(1<<LED9);
    DDRB=(1<<LED6)|(1<<LED7);
    TCCR0A=0x02; //ctc mode
    OCR0A=0;
    TCCR0B=4;
    TIMSK0=0x2; //int on compare a
}

//Measures the photo-transistor
short measureFoto() {
    short m;
    ADMUX=0x40; //measure pc0 against vcc
    ADCSRA=0xC3; //go measure
    while (ADCSRA&0x40) ; //wait for adc end
    m=ADC;
    return m;
}

//Turn the indicated led on or off.
void setled(char led, char on) {
    if (on) {
	if (led==0) PORTD&=~(1<<LED1);
	if (led==1) PORTD&=~(1<<LED2);
	if (led==2) PORTD&=~(1<<LED3);
	if (led==3) PORTD&=~(1<<LED4);
	if (led==4) PORTD&=~(1<<LED5);
	if (led==5) PORTB&=~(1<<LED6);
	if (led==6) PORTB&=~(1<<LED7);
	if (led==7) PORTD&=~(1<<LED8);
	if (led==8) PORTD&=~(1<<LED9);
    } else {
	if (led==0) PORTD|=(1<<LED1);
	if (led==1) PORTD|=(1<<LED2);
	if (led==2) PORTD|=(1<<LED3);
	if (led==3) PORTD|=(1<<LED4);
	if (led==4) PORTD|=(1<<LED5);
	if (led==5) PORTB|=(1<<LED6);
	if (led==6) PORTB|=(1<<LED7);
	if (led==7) PORTD|=(1<<LED8);
	if (led==8) PORTD|=(1<<LED9);
    }

}

//Make the interrupt routine recalibrate completely. No button should be
//pressed now!
void resetCal() {
    calButtons=1;
    while(calButtons!=0);
}

#define FOTO_TRESH 5
#define FOTO_DELAY 100
#define CONSEC_SAMPLES 4

//Measure the photo-response to lighting of the leds. Deciding if a button should
//be on or off is done by vote: every time the routine thinks the button is on,
//it'll increade buttons[x] by one, if not it'll decrease it two times. If the
//total is larger than CONSEC_SAMPLES, the rest of the program will be told the
//button was on.
void getPhoto() {
    static char buttons[9];
    //Because of smudges etc, every led will return a small amount of light, even when 
    //not pressed. photoBase contains this amount so we can compensate for it.
    static short photoBase[9];
    short buttonsNew;
    short meas[9];
    short base[9];
    short d;
    int x;
    
    buttonsNew=0;

    //switch off leds
    for (x=0; x<9; x++) setled(x,0);
    _delay_us(FOTO_DELAY*2);
    //Go measure dark/lit phototransistor response for each led.
    //Darkness measurements will go into base[x], light into meas[x].
    for (x=0; x<9; x++) {
	setled(x,0);
	_delay_us(FOTO_DELAY);
	base[x]=measureFoto();
	setled(x,1);
	_delay_us(FOTO_DELAY);
	meas[x]=measureFoto();
	setled(x,0);
	_delay_us(FOTO_DELAY);
	base[x]+=measureFoto();
	base[x]/=2;
    }

    //See if any button was pressed.
    for (x=0; x<9; x++) {
	//Get the measurement and see how much we're above 'normal', unpressed values.
	d=meas[x]-(base[x]+photoBase[x]);
	if (d>FOTO_TRESH) { //Do we think the button is pressed?
	    //Yes.
    	    buttons[x]++; 
    	} else {
    	    //Nope.
    	    if (buttons[x]!=0) buttons[x]--;
    	    if (buttons[x]!=0) buttons[x]--;
    	}
    	//Generate actual button press if buttons[x] is high enough
    	if (buttons[x]>CONSEC_SAMPLES) buttonsNew|=(1<<x);
    	//clip buttons[x] so its value won't go through the roof
    	if (buttons[x]>CONSEC_SAMPLES+3) buttons[x]=CONSEC_SAMPLES+3;
    }

    //If calButtons is on, we can assume no button is pressed at all and
    //all light returned from the led to the phototransistor is because of
    //smudges etc.
    if (calButtons) {
	buttonsNew=0;
	for (x=0; x<9; x++) {
	    photoBase[x]=meas[x]-base[x];
	    if (photoBase[x]<0) photoBase[x]=0;
	}
    }

    //If buttons are pressed, assume the rest really is unpressed and partly calibrate
    //these buttons.
    if (buttonsNew!=0) {
	for (x=0; x<9; x++) {
	    if ((buttonsNew&(1<<x))==0) photoBase[x]=(photoBase[x]/2)+(meas[x]-base[x]);
	    if (photoBase[x]<0) photoBase[x]=0;
	}
    }
    //return buttons.
    pressedButtons=buttonsNew;
}

//This is the timer routine. It will be called every 1, 2, 4, 8 etc
//cycles because it sets its own ocra value: it is an implementation
//of binary code modulation as described here:
//http://www.batsocks.co.uk/readme/art_bcm_1.htm
ISR(TIMER0_COMPA_vect) {
    static unsigned char bitfield;
    static char realLedIntensity[9];
    int x;
    
    bitfield=bitfield<<1;
    
    if (bitfield==0) { //bit fell out, time to do some maintenance.
	//This is a nice time to do the button-reading procedure.
	getPhoto();
	calButtons=0;
	
    	//Copy intensity settings now.
	for (x=0; x<9; x++) {
	    realLedIntensity[x]=ledIntensity[x];
	}
	bitfield=1;
	//Reset timer because we might have been a while.
	TIFR0|=2; //reset pending int
	TCNT0=0;
    }

    OCR0A=bitfield;
    for (x=0; x<9; x++) {
	if (realLedIntensity[x]&bitfield) setled(x,1); else setled(x,0);
    }
}

